/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/vfs.h>


#include "io.h"
#include "misc.h"
#include "reiserfs_lib.h"
#include "../version.h"



#define print_usage_and_exit() die ("Usage: %s [-b block-to-print][-idc] device\n\
-i Causes to print all items of a leaf\n\
-d                 content of directory items\n\
-c                 content of direct items\n\
-m                 bitmap blocks\n", argv[0]);



#if 1
struct reiserfs_fsstat {
    int nr_internals;
    int nr_leaves;
    int nr_files;
    int nr_directories;
    int nr_unformatted;
} g_stat_info;
#endif

/*
 *  modes
 */
#define DO_DUMP 1  /* not a real dump, just dumping contents of tree nodes */
#define DO_PACK 2  /* helps debugging fsck */
#define DO_CORRUPT 3 /* used to make filesystem corruption and then test fsck */
#define DO_SCAN 4
#define DO_SCAN_FOR_NAME 5
#define DO_RECOVER 6
#define DO_TEST 7

int mode = DO_DUMP;

/*
 *  options
 */
int opt_print_regular_file_content = 0;/* -c */
int opt_print_details = 0;	/* -d */
int opt_print_leaf_items = 0;		/* -i */
int opt_print_objectid_map = 0;	/* -o */
int opt_print_block_map = 0;		/* -m */
/* when you want print one block specify -b # */
int opt_block_to_print = -1;
/* when you want to corrupt block specify -C # */
int opt_block_to_corrupt = -1;
int opt_pack = 0;	/* -P will produce output that should be |gzip -c > whatever.gz */
			/* -p will calculate number of bytes needed to transfer the partition */
int opt_print_journal;
int opt_pack_all = 0;


int print_mode (void)
{
    int mode = 0;

    if (opt_print_leaf_items == 1)
	mode |= PRINT_LEAF_ITEMS;
    if (opt_print_details == 1)
	mode |= (PRINT_LEAF_ITEMS | PRINT_ITEM_DETAILS);
    if (opt_print_regular_file_content == 1)
	mode |= (PRINT_LEAF_ITEMS | PRINT_DIRECT_ITEMS);
    return mode;
}


static void print_disk_tree (reiserfs_filsys_t fs, int block_nr)
{
    struct buffer_head * bh;

    bh = bread (fs->s_dev, block_nr, fs->s_blocksize);
    if (!bh) {
	die ("Could not read block %d\n", block_nr);
    }
    if (is_internal_node (bh)) {
	int i;
	struct disk_child * dc;

	g_stat_info.nr_internals ++;
	print_block (fs, bh, print_mode (), -1, -1);
      
	dc = B_N_CHILD (bh, 0);
	for (i = 0; i <= B_NR_ITEMS (bh); i ++, dc ++)
	    print_disk_tree (fs, dc->dc_block_number);
      
    } else if (is_leaf_node (bh)) {
	g_stat_info.nr_leaves ++;
	print_block (fs, bh, print_mode (), -1, -1);
    } else {
	print_block (fs, bh, print_mode (), -1, -1);
	die ("print_disk_tree: bad block type");
    }
    brelse (bh);
}



void pack_one_block (reiserfs_filsys_t fs, unsigned long block);
static void print_one_block (reiserfs_filsys_t fs, int block)
{
    struct buffer_head * bh;
    
    if (test_bit (block % (fs->s_blocksize * 8), 
		  SB_AP_BITMAP (fs)[block / (fs->s_blocksize * 8)]->b_data))
	fprintf (stderr, "%d is used in true bitmap\n", block);
    else
	fprintf (stderr, "%d is free in true bitmap\n", block);
    
    bh = bread (fs->s_dev, block, fs->s_blocksize);
    if (!bh) {
	printf ("print_one_block: bread fialed\n");
	return;
    }

    if (opt_pack) {
	pack_one_block (fs, bh->b_blocknr);
	brelse (bh);
	return;
    }

    if (who_is_this (bh->b_data, fs->s_blocksize) != THE_UNKNOWN)
	print_block (fs, bh, PRINT_LEAF_ITEMS | PRINT_ITEM_DETAILS | 
		     (opt_print_regular_file_content == 1 ? PRINT_DIRECT_ITEMS : 0), -1, -1);
    else
	printf ("Looks like unformatted\n");
    brelse (bh);
    return;
}


static void corrupt_clobber_hash (char * name, struct item_head * ih, 
				  struct reiserfs_de_head * deh)
{
    printf ("\tCorrupting deh_offset of entry \"%s\" of [%u %u]\n", name,
	    ih->ih_key.k_dir_id, ih->ih_key.k_objectid);
    deh->deh_offset = 700;
}


static void corrupt_delete_entry (reiserfs_filsys_t fs,
				  char * name,
				  struct item_head * ih,
				  struct reiserfs_de_head * deh)
{
    struct key key;

    printf ("\tDeleting entry \"%s\" of [%u %u] - ", name,
	    ih->ih_key.k_dir_id, ih->ih_key.k_objectid);
	    
    key.k_dir_id = ih->ih_key.k_dir_id;
    key.k_objectid = ih->ih_key.k_objectid;
    key.u.k_offset_v1.k_offset = deh->deh_offset;
    key.u.k_offset_v1.k_uniqueness = 500;
    if (reiserfs_remove_entry (fs, &key))
	printf ("failed\n");
    else
	printf ("ok\n");
}


/* this reads list of desired corruptions from stdin and perform the
   corruptions. Format of that list:

   C name objectid - 'C'ut entry 'name' from directory item with 'objectid'
   H name objectid - clobber 'H'hash of entry 'name' of directory 'objectid'
   O item_num      - destroy item 'O'rder - make 'item_num'-th to have key bigger than 'item_num' + 1-th item
*/
static void do_corrupt_one_block (reiserfs_filsys_t fs, int block)
{
    struct buffer_head * bh;
    int i, j;
    struct item_head * ih;
    int item_num;
    char * line = 0;
    int n = 0;
    char code, name [100];
    __u32 objectid;
    

    if (test_bit (block % (fs->s_blocksize * 8), 
		  SB_AP_BITMAP (fs)[block / (fs->s_blocksize * 8)]->b_data))
	fprintf (stderr, "%d is used in true bitmap\n", block);
    else
	fprintf (stderr, "%d is free in true bitmap\n", block);
    
    bh = bread (fs->s_dev, block, fs->s_blocksize);
    if (!bh) {
	printf ("corrupt_one_block: bread fialed\n");
	return;
    }

    if (who_is_this (bh->b_data, fs->s_blocksize) != THE_LEAF) {
	printf ("Can not corrupt not a leaf node\n");
	brelse (bh);
	return;
    }

    printf ("Corrupting block %lu..\n", bh->b_blocknr);

    while (getline (&line, &n, stdin) != -1) {
	switch (line[0]) {
	case '#':
	case '\n':
	    continue;
	case 'C':
	case 'H':
	    if (sscanf (line, "%c %s %u\n", &code, name, &objectid) != 3) {
		printf ("Wrong format \'%c\'\n", line [0]);
		continue;
	    }
	    break;

	case 'O':
	    if (sscanf (line, "%c %d\n", &code, &item_num) != 2) {
		printf ("Wrong format \'O\'\n");
		continue;
	    }
	}

	ih = B_N_PITEM_HEAD (bh, 0);
	for (i = 0; i < node_item_number (bh); i ++, ih ++) {
	    struct reiserfs_de_head * deh;

	    if (code == 'O' && i == item_num) {
		/* destroy item order */
		struct key * key;
		if (i == node_item_number (bh) - 1) {
		    printf ("can not destroy order\n");
		    continue;
		}
		key = &(ih + 1)->ih_key;
		ih->ih_key.k_dir_id = key->k_dir_id + 1;
		mark_buffer_dirty (bh);
	    }

	    if (ih->ih_key.k_objectid != objectid || !is_direntry_ih (ih))
		continue;

	    deh = B_I_DEH (bh, ih);

	    for (j = 0; j < ih_entry_count (ih); j ++, deh ++) {
		/* look for proper entry */
		if (name_length (ih, deh, j) != strlen (name) ||
		    strncmp (name, name_in_entry (deh, j), strlen (name)))
		    continue;

		/* ok, required entry found, make a corruption */
		switch (code) {
		case 'C': /* cut entry */
		    corrupt_delete_entry (fs, name, ih, deh);
		    if (!B_IS_IN_TREE (bh)) {
			printf ("NOTE: block is deleted from the tree\n");
			exit (0);
		    }
		    goto cont;
		    break;
		case 'H': /* clobber hash */
		    corrupt_clobber_hash (name, ih, deh);
		    goto cont;
		    break;
		default:
		    printf ("Unknown command found\n");
		}
		mark_buffer_dirty (bh);
	    }
	}
    cont:
    }
    free (line);
    printf ("Done\n");
    brelse (bh);
    return;
}


/* this reads stdin and recover file of given key:  */
/* the input has to be in the follwong format:
   K dirid objectid
   N name
   B blocknumber
   ..
   then recover_file will read every block, look there specified file and put it into
*/
static void do_recover (reiserfs_filsys_t fs)
{
    char name [100];
    char * line = 0;
    int n = 0;
    int fd;
    struct key key = {0, 0, };
    struct buffer_head * bh;
    struct item_head * ih;
    unsigned long block;
    char code;
    loff_t recovered = 0;
    int i, j;
    reiserfs_bitmap_t bitmap;
    int used, not_used;

    bitmap = reiserfs_create_bitmap (SB_BLOCK_COUNT (fs));
    reiserfs_fetch_disk_bitmap (bitmap, fs);
    /* we check how many blocks recoverd items point to are free or used */
    used = 0;
    not_used = 0;

    fd = 0;
    while (getline (&line, &n, stdin) != -1) {
	if (line [0] == '#' || line [0] == '\n')
	    continue;
	switch (line [0]) {
	case 'K':
	    /* get a key of file which is to be recovered */
	    if (sscanf (line, "%c %u %u\n", &code, &key.k_dir_id, &key.k_objectid) != 3) {
		die ("recover_file: wrong input K format");
	    }
	    printf ("Recovering file (%u, %u)\n", key.k_dir_id, key.k_objectid);
	    break;

	case 'N':
	    /* get a file name */
	    recovered = 0;
	    if (sscanf (line, "%c %s\n", &code, name) != 2) {
		die ("recover_file: wrong input N format");
	    }
	    fd = open (name, O_RDWR | O_CREAT | O_EXCL, 0644);
	    if (fd == -1)
		die ("recover_file: could not create file %s: %s",
		     name,strerror (errno));
	    printf ("Recovering file %s..\n", name);
	    break;

	case 'B':
	    if (!fd)
		die ("recover_file: file name is not specified");
	    if (sscanf (line, "%c %lu\n", &code, &block) != 2) {
		die ("recover_file: wrong input B format");
	    }
	    bh = bread (fs->s_dev, block, fs->s_blocksize);
	    if (!bh) {
		printf ("reading block %lu failed\n", block);
		continue;
	    }

	    printf ("working with block %lu..\n", block);

	    ih = B_N_PITEM_HEAD (bh, 0);
	    for (i = 0; i < node_item_number (bh); i ++, ih ++) {
		__u32 * indirect;
		struct buffer_head * tmp_bh;

		if (!is_indirect_ih (ih) || key.k_dir_id != ih->ih_key.k_dir_id ||
		    key.k_objectid != ih->ih_key.k_objectid)
		    continue;

		indirect = (__u32 *)B_I_PITEM (bh, ih);
		for (j = 0; j < I_UNFM_NUM (ih); j ++) {
		    block = le32_to_cpu (indirect [j]);
		    if (!block)
			continue;
		    tmp_bh = bread (fs->s_dev, block, fs->s_blocksize);
		    if (!tmp_bh) {
			printf ("reading block %Lu failed\n", (loff_t)block * fs->s_blocksize);
			continue;
		    }
		    if (reiserfs_llseek (fd, get_offset (&ih->ih_key) + j * fs->s_blocksize - 1,
					 SEEK_SET) == (loff_t)-1) {
			printf ("llseek failed to pos %Ld\n", (loff_t)block * fs->s_blocksize);
			brelse (tmp_bh);
			continue;
		    }
		    if (reiserfs_bitmap_test_bit (bitmap, block))
			used ++;
		    else
			not_used ++;
		    /*printf ("block of file %Ld gets block %lu\n",
		      (get_offset (&ih->ih_key) - 1) / fs->s_blocksize + j, block);*/
		    if (write (fd, tmp_bh->b_data, tmp_bh->b_size) != tmp_bh->b_size) {
			printf ("write failed to pos %Ld\n", (loff_t)block * fs->s_blocksize);
			brelse (tmp_bh);
			continue;
		    }
		    recovered += fs->s_blocksize;
		    brelse (tmp_bh);
		}
	    }
	    brelse (bh);
	    break;
	}
    }
    printf ("recover_file: %Ld bytes recovered of file %s, key %u %u, %d blocks are free and %d are used\n",
	    recovered, name, key.k_dir_id, key.k_objectid, not_used, used);
}


static char * parse_options (int argc, char * argv [])
{
    int c;
    char * tmp;
  
    while ((c = getopt (argc, argv, "b:C:icdmoMpPaAjsnrt")) != EOF) {
	switch (c) {
	case 'b':	/* print a single node */
	    opt_block_to_print = strtol (optarg, &tmp, 0);
	    if (*tmp)
		die ("parse_options: bad block size");
	    break;
	case 'C':
	    mode = DO_CORRUPT;
	    opt_block_to_corrupt = strtol (optarg, &tmp, 0);
	    if (*tmp)
		die ("parse_options: bad block size");
	    break;
	    
	case 'p':
	case 'P':
	    mode = DO_PACK;
	    break;

	case 'i':	/* print items of a leaf */
	    opt_print_leaf_items = 1; break;

	case 'd':	/* print directories */
	    opt_print_details = 1; break;

	case 'c':	/* print contents of a regular file */
	    opt_print_regular_file_content = 1; break;

	case 'o':	/* print a objectid map */
	    opt_print_objectid_map = 1; break;

	case 'm':	/* print a block map */
	    opt_print_block_map = 1;  break;
	case 'M':	/* print a block map with details */
	    opt_print_block_map = 2;  break;
	case 'j':
	    opt_print_journal = 1; break; /* print journal */
	    
	case 's':
	    mode = DO_SCAN; break; /* read the device and print what reiserfs blocks were found */

	case 'n':
	    mode = DO_SCAN_FOR_NAME; break;

	case 'r':
	    mode = DO_RECOVER; break;

	case 't':
	    mode = DO_TEST; break;
	}
    }
    if (optind != argc - 1)
	/* only one non-option argument is permitted */
	print_usage_and_exit();
  
    return argv[optind];
}



/* print all valid transactions and found dec blocks */
static void print_journal (struct super_block * s)
{
    struct buffer_head * d_bh, * c_bh;
    struct reiserfs_journal_desc * desc ;
    struct reiserfs_journal_commit *commit ;
    int end_journal;
    int start_journal;
    int i;
    int first_desc_block = 0;
    int wrapped = 0;
    int valid_transactions = 0;

    start_journal = SB_JOURNAL_BLOCK (s);
    end_journal = start_journal + JOURNAL_BLOCK_COUNT;
    reiserfs_warning2 (stderr, "Start scanning from %d\n", start_journal);

    d_bh = 0;
    desc = 0;
    for (i = start_journal; i < end_journal; i ++) {
	d_bh = bread (s->s_dev, i, s->s_blocksize);
	if (who_is_this (d_bh->b_data, d_bh->b_size) == THE_JDESC) {
	    int commit_block;

	    if (first_desc_block == 0)
		/* store where first desc block found */
		first_desc_block = i;

	    print_block (s, d_bh); /* reiserfs_journal_desc structure will be printed */
	    desc = (struct reiserfs_journal_desc *)(d_bh->b_data);

	    commit_block = d_bh->b_blocknr + desc->j_len + 1;
	    if (commit_block >= end_journal) {
		reiserfs_warning2 (stderr, "-- wrapped?");
		wrapped = 1;
		break;
	    }

	    c_bh = bread (s->s_dev, commit_block, s->s_blocksize);
	    commit = bh_commit (c_bh);
	    if (does_desc_match_commit (desc, commit)) {
		reiserfs_warning2 (stderr, "commit block %d (trans_id %ld, j_len %ld) does not match\n", commit_block,
				   commit->j_trans_id, commit->j_len);
		brelse (c_bh) ;
		brelse (d_bh);
		continue;
	    }

	    valid_transactions ++;
	    reiserfs_warning2 (stderr, "(commit block %d) - logged blocks (", commit_block);
#if 0
	    for (j = 0; j < desc->j_len; j ++) {
		unsigned long block;

		if (j < JOURNAL_TRANS_HALF)
		    block = le32_to_cpu (desc->j_realblock[j]);
		else
		    block = le32_to_cpu (commit->j_realblock[i - JOURNAL_TRANS_HALF]);
			
		if (not_journalable (s, block))
		    reiserfs_warning2 (stderr, " xxxx");
		else {
		    reiserfs_warning2 (stderr, " %ld", desc->j_realblock[j]);
		    if (block_of_bitmap (s, desc->j_realblock[j]))
			reiserfs_warning2 (stderr, "(bmp)");
		}
		if (j && (j + 1) % 10 == 0)
		    reiserfs_warning2 (stderr, "\n");
	    }
#endif
	    reiserfs_warning2 (stderr, ")\n");
	    i += desc->j_len + 1;
	    brelse (c_bh);
	}
	brelse (d_bh);
    }
    
    if (wrapped) {
	c_bh = bread (s->s_dev, first_desc_block - 1, s->s_blocksize);
	commit = bh_commit (c_bh);
	if (does_desc_match_commit (desc, commit)) {
	    reiserfs_warning2 (stderr, "No! commit block %d (trans_id %ld, j_len %ld) does not match\n",
			       first_desc_block - 1, commit->j_trans_id, commit->j_len);
	} else {
	    reiserfs_warning2 (stderr, "Yes! (commit block %d) - logged blocks (\n", first_desc_block - 1);
#if 0
	    for (j = 0; j < desc->j_len; j ++) {
		unsigned long block;

		if (j < JOURNAL_TRANS_HALF)
		    block = le32_to_cpu (desc->j_realblock[j]);
		else
		    block = le32_to_cpu (commit->j_realblock[i - JOURNAL_TRANS_HALF]);
			
		if (not_journalable (s, block))
		    reiserfs_warning2 (stderr, " xxxx");
		else {
		    reiserfs_warning (" %ld", desc->j_realblock[j]);
		    if (block_of_bitmap (s, desc->j_realblock[j]))
			reiserfs_warning2 (stderr, "(bmp)");
		}
	    }
#endif
	    reiserfs_warning2 (stderr, "\n");
	}
	brelse (c_bh) ;
	brelse (d_bh);
    }

    reiserfs_warning2 (stderr, "%d valid transactions found\n", valid_transactions);

    {
	struct buffer_head * bh;
	struct reiserfs_journal_header * j_head;

	bh = bread (s->s_dev, SB_JOURNAL_BLOCK (s) + rs_journal_size (s->s_rs),
		    s->s_blocksize);
	j_head = (struct reiserfs_journal_header *)(bh->b_data);

	reiserfs_warning2 (stderr, "#######################\nJournal header:\n"
			   "j_last_flush_trans_id %ld\n"
			   "j_first_unflushed_offset %ld\n"
			   "j_mount_id %ld\n", j_head->j_last_flush_trans_id, j_head->j_first_unflushed_offset,
			   j_head->j_mount_id);
	brelse (bh);
    }
}


void pack_partition (reiserfs_filsys_t fs);

static void do_pack (reiserfs_filsys_t fs)
{
    if (opt_block_to_print != -1)
	pack_one_block (fs, opt_block_to_print);
    else
	pack_partition (fs);
	
}

/* FIXME: statistics does not work */
static void do_dump_tree (reiserfs_filsys_t fs)
{
    if (opt_block_to_print != -1) {
	print_one_block (fs, opt_block_to_print);
	return;
    }

    print_block (fs, SB_BUFFER_WITH_SB (fs));
    
    if (opt_print_journal)
	print_journal (fs);
    
    if (opt_print_objectid_map == 1)
	print_objectid_map (fs);
    
    if (opt_print_block_map)
	print_bmap (fs, opt_print_block_map == 1 ? 1 : 0);
    
    if (opt_print_regular_file_content || opt_print_details ||
	opt_print_leaf_items) {
	print_disk_tree (fs, SB_ROOT_BLOCK (fs));
	
	/* print the statistic */
	printf ("File system uses %d internal + %d leaves + %d unformatted nodes = %d blocks\n",
		g_stat_info.nr_internals, g_stat_info.nr_leaves, g_stat_info.nr_unformatted, 
		g_stat_info.nr_internals + g_stat_info.nr_leaves + g_stat_info.nr_unformatted);
    }
}

FILE * log;

static void look_for_key (struct buffer_head * bh, struct key * key)
{
    int i;
    struct item_head * ih;

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih ++) {
	if (ih->ih_key.k_dir_id == key->k_dir_id &&
	    ih->ih_key.k_objectid == key->k_objectid) {
	    fprintf (log, "block %lu has item of file %u %u (item %d)\n",
		     bh->b_blocknr, key->k_dir_id, key->k_objectid, i);fflush (log);
	    return;
	}
    }
    return;
}


static void look_for_name (struct buffer_head * bh, char * name)
{
    int i, j;
    struct item_head * ih;
    struct reiserfs_de_head * deh;
    int namelen;
    char * p;

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih ++) {
	if (!is_direntry_ih (ih))
	    continue;
	deh = B_I_DEH (bh, ih);
	for (j = 0; j < ih_entry_count (ih); j ++, deh ++) {
	    p = name_in_entry (deh, j);
	    namelen = name_length (ih, deh, j);
	    if (namelen == strlen (name) && !strncmp (name, p, namelen)) {
		fprintf (log, "block %lu, item %d, entry %d is %s\n", bh->b_blocknr, i, j, name);fflush (log);
	    }
	}
    }
    return;
}




static void do_scan (reiserfs_filsys_t fs)
{
    unsigned long i;
    struct buffer_head * bh;
    int type;
    char * answer = 0;
    size_t n = 0;
    struct key key = {0, 0, };
    unsigned long done;

    log = fopen ("scan.log", "w+");

    if (mode == DO_SCAN_FOR_NAME) {
	printf ("What name do you want to look for?");
	getline (&answer, &n, stdin);
	answer [strlen (answer) - 1] = 0;
	printf ("Looking for name \"%s\"..\n", answer);
	key.k_dir_id = 1;
    } else {
	printf ("What key do you want to find: dirid?");
	getline (&answer, &n, stdin);
	key.k_dir_id = atoi (answer);
	printf ("objectid?");
	getline (&answer, &n, stdin);
	key.k_objectid = atoi (answer);
	printf ("looking for (%u %u)\n", key.k_dir_id, key.k_objectid);
    }

    done = 0;
    for (i = 0; i < SB_BLOCK_COUNT (fs); i ++) {
	bh = bread (fs->s_dev, i, fs->s_blocksize);
	if (!bh) {
	    printf ("could not read block %lu\n", i);
	    continue;
	}
	type = who_is_this (bh->b_data, bh->b_size);
	switch (type) {
	case THE_JDESC:
	    if (!key.k_dir_id)
		printf ("block %lu is journal descriptor\n", i);
	    break;
	case THE_SUPER:
	    if (!key.k_dir_id)
		printf ("block %lu is reiserfs super block\n", i);
	    break;
	case THE_INTERNAL:
	    if (!key.k_dir_id)
		printf ("block %lu is reiserfs internal node\n", i);
	    break;
	case THE_LEAF:
	    if (mode == DO_SCAN_FOR_NAME) {
		look_for_name (bh, answer);
	    } else if (key.k_dir_id) {
		look_for_key (bh, &key);
	    } else {
		printf ("block %lu is reiserfs leaf node\n", i);
	    }
	    break;
	}
	brelse (bh);
	print_how_far (&done, SB_BLOCK_COUNT (fs));
    }
}


/*############## things which are needed to make test section compilable */

static char g_bad_name [4096];

char * bad_name (char * name, int namelen)
{
    memcpy (g_bad_name, name, namelen);
    g_bad_name [namelen] = 0;
    return g_bad_name;
}

#define fsck_log(fmt, list...) reiserfs_warning2 (stderr, fmt, ## list)
#define fsck_user_confirmed(fs,a,q,default) user_confirmed (a,q)
#define fsck_mode(fs) 1
#define FSCK_ZERO_FILES 0
reiserfs_bitmap_t f;
#define fsck_new_bitmap(fs) f
int tmp_zeroed;
#define stat_wrong_pointer_found(fs) 1
#define was_block_used(u) 1
#define register_unfm(u) 0
#define pass0_mark_leaf(bh) 0
#define stat_pass0_leaf_corected(fs) 0

/*############## code to be debugged ############################################*/

/* 'upper' item is correct if 'upper + 2' exists and its key is greater than
   key of 'upper' */
static int upper_correct (struct buffer_head * bh, struct item_head * upper,
			  int upper_item_num)
{
    if (upper_item_num + 2 < B_NR_ITEMS (bh)) {
	if (comp_keys (&upper->ih_key, &(upper + 2)->ih_key) != -1)
	    /* item-num's item is out of order of order */
	    return 0;
	return 1;
    }
    
    /* there is no item above the "bad pair" */
    return 2;
}


/* 'lower' item is correct if 'lower - 2' exists and its key is smaller than
   key of 'lower' */
static int lower_correct (struct buffer_head * bh, struct item_head * lower,
			  int lower_item_num)
{
    if (lower_item_num - 2 >= 0) {
	if (comp_keys (&(lower - 2)->ih_key, &lower->ih_key) != -1)
	    return 0;
	return 1;
    }
    return 2;
}


#if 0
/* for now: recover only "." and "..". Other improperly hashed entries get
   deleted (fimxe that)*/
static int fix_entry (reiserfs_filsys_t fs, struct buffer_head * bh,
		      struct item_head * ih, char * name, int namelen,
		      struct reiserfs_de_head * deh, 
		      int item_num, int entry_num)
{
    if (get_offset (&ih->ih_key) == DOT_OFFSET && entry_num < 2) {
	/* either '.' or '..' is corrupted */
	fsck_log ("pass0: block %lu: %d-th name \"%s\" in directory %k "
		  "is not hashed properly (fixed)\n",
		  bh->b_blocknr, entry_num, bad_name (name, namelen), &ih->ih_key);
	name[0] = '.';
	if (entry_num == 0) {
	    deh->deh_offset = cpu_to_le32 (DOT_OFFSET);
	} else {
	    name[1] = '.';
	    deh->deh_offset = cpu_to_le32 (DOT_DOT_OFFSET);
	}
	return 0;
    }
    fsck_log ("pass0: block %lu: %d-th name \"%s\" in directory %k "
	      "is not hashed properly (will be removed)\n",
	      bh->b_blocknr, entry_num, bad_name (name, namelen), &ih->ih_key);

    /*cut_entry (fs, bh, item_num, entry_num);*/
    return 1;
}


/* returns one if could not recover directory item. For now it only recovers
   directory item if locations are broken. FIXME: check how that will work on
   decently broken directory entry */
static int correct_direntry (reiserfs_filsys_t fs, struct buffer_head * bh,
			     struct item_head * ih, int item_num)
{
    int entry_num;
    char * name;
    int namelen;
    struct reiserfs_de_head * deh = (struct reiserfs_de_head *)B_I_PITEM (bh, ih);
    struct reiserfs_de_head * deh1;
    __u32 prev_offset = 0;
    __u16 prev_location = ih_item_len (ih);
    int min_entry_size = 1;/* we have no way to understand whether the
                              filesystem were created in 3.6 format or
                              converted to it. So, we assume that minimal name
                              length is 1 */
    int dirty;
    __u16 proper_state;
    int minimal_location;




    if (ih_item_len (ih) / (DEH_SIZE + min_entry_size) < ih_entry_count (ih))
	/* entry count is too big */
	return 1;
    
    deh1 = deh;
    dirty = 0;
    proper_state = 0;
    set_bit (DEH_Visible, &proper_state);    
    minimal_location = DEH_SIZE * ih_entry_count (ih);

    /* in this loop we only mark entries fro deletion - fixme: this is
       probably does not work when two or more neighboring entries are broken */
    for (entry_num = 0; entry_num < ih_entry_count (ih); entry_num ++, deh ++) {

	/* make sure that entries are sorted properly */
	if (deh_offset (deh) <= prev_offset) {
	    /* fixme: we do not recover from this yet */
	    return 1;
	}
	prev_offset = deh_offset (deh);


	/* check location of the current entry */
	if (deh_location (deh) >= prev_location ||
	    deh_location (deh) < minimal_location) {
	    /* location looks suspicious, skip the name */
	    fsck_log ("%d-th entry has bad location %d\n", entry_num, deh_location (deh));
	    deh->deh_location = 0;
	    prev_location = 0xffff;
	    continue;
	}

	if (prev_location == 0xffff) {
	    int i, bound;
	    struct reiserfs_de_head * tmp;
	    /* previous name was skipped, so we do not know where current name
               ends. Try to find it */

	    bound = ih_item_len (ih);
	    for (tmp = deh - 1; tmp >= deh1; tmp --)
		if (deh_location (tmp)) {
		    bound = deh_location (tmp) - deh_location (deh);
		    break;
		}

	    /* fixme: we should try only down to the nearest valid entry */
	    name = name_in_entry (deh, entry_num);
	    for (i = 1; i < bound; i ++) {
		if (is_properly_hashed (fs, name, i, deh_offset (deh)))
		    break;
	    }
	    if (i == bound) {
		/* ok, make deh_locations such that directory will be looking
                   like a directory */
		struct reiserfs_de_head * tmp;

		fsck_log ("pass0: block %lu: could not find the end of %d-th name in item %H\n",
			  bh->b_blocknr, entry_num, ih);
		mark_de_bad (deh);
		/*deh->deh_location = 0;
		  prev_location = 0xffff;*/

		/* set deh_location to the nearest dehs with zeroed locations */
		for (tmp = deh - 1; tmp >= deh1; tmp --) {
		    if (deh_location (tmp)) {
			break;
		    }
		    tmp->deh_location = (tmp + 1)->deh_location + 1;
		}
		
		prev_location = deh_location (deh);
		continue;		
	    }

	    /* ok, set deh_location of previous names */
	    if (!entry_num)
		die ("correct_direntry: wrong prev location");
	    fsck_log ("correct_direntry: ok, found the end of %d-th name \"%s\"\n",
		      entry_num, bad_name (name, i));
	    (deh - 1)->deh_location = cpu_to_le16 (deh_location (deh) + i);

	    /* set deh_location to the nearest dehs with zeroed locations */
	    for (tmp = deh - 2; tmp >= deh1; tmp --) {
		if (deh_location (tmp)) {
		    break;
		}
		tmp->deh_location = (tmp + 1)->deh_location + 1;
	    }
	    /* ok, things restored */
	    prev_location = deh_location (deh);
	    continue;
	}


	prev_location = deh_location (deh);

	/* fix state */
	if (deh_state (deh) != proper_state) {
	    deh->deh_state = cpu_to_le16 (proper_state);
	    dirty = 1;
	}

	namelen = name_length (ih, deh, entry_num);
	name = name_in_entry (deh, entry_num);
	if (!is_properly_hashed (fs, name, namelen, deh_offset (deh))) {
	    /* correct "." and ".." and delele other corrupted name */
	    if (fix_entry (fs, bh, ih, name, namelen, deh, item_num, entry_num))
		mark_de_bad (deh);
	    else
		/* '.' or '..' recovered */;
	}
    }

    /* delete all entries which has been marked unrecoverable */
 start_again:
    deh = B_I_DEH (bh, ih);
    for (entry_num = 0; entry_num < ih_entry_count (ih); entry_num ++, deh ++) {
	if (de_bad (deh)) {
	    fsck_log ("correct_direntry: block %lu, item %H, %d-th entry removed\n",
		      bh->b_blocknr, ih, entry_num);
	    cut_entry (fs, bh, item_num, entry_num);
	    goto start_again;
	}
    }

    if (dirty)
	mark_buffer_dirty (bh);
#if 0
    if (is_bad_directory (ih, B_I_PITEM (bh, ih), fs->s_dev, fs->s_blocksize)) {
	die ("directory item was not recovered");
    }
#endif
    return 0;
}
#endif

static int prob_name (reiserfs_filsys_t fs,
		      char ** name, int max_len, __u32 masked_hash)
{
    int start; /* */
    int len;

    for (start = 0; start < max_len; start ++) {
	for (len = 0; len < max_len - start; len ++) {
	    if (is_properly_hashed (fs, *name + start, len + 1, masked_hash)) {
		*name = *name + start;
		return len + 1;
	    }
	}
    }
    return 0;
}


/* recover what can be deleted (w/ reasonable efforts) and delete the rest */
int verify_directory_item (reiserfs_filsys_t fs, struct buffer_head * bh,
			   int item_num)
{
    struct item_head * ih;
    char * item;
    struct reiserfs_de_head * deh;
    char * entries, * end;
    __u32 masked_offset;
    int total_entry_len, name_len;
    int i;
    int bad_entries; /* how many bad neighboring entries */
    int dirty;

    dirty = 0;

    ih = B_N_PITEM_HEAD (bh, item_num);
    item = B_I_PITEM (bh,ih);
    deh = (struct reiserfs_de_head *)item;

    if (get_offset (&ih->ih_key) == DOT_OFFSET) {
	/* first item of directory - make sure that "." and ".." are in place */
	if (deh_offset (deh) != DOT_OFFSET || name_in_entry (deh, 0)[0] != '.') {
	    deh->deh_offset = cpu_to_le32 (DOT_OFFSET);
	    name_in_entry (deh, 0)[0] = '.';
	    dirty = 1;
	}
	if (deh_offset (deh + 1) != DOT_DOT_OFFSET ||
	    name_in_entry (deh + 1, 1)[0] != '.' || name_in_entry (deh + 1, 1)[1] != '.') {
	    (deh + 1)->deh_offset = cpu_to_le32 (DOT_DOT_OFFSET);
	    name_in_entry (deh + 1, 1)[0] = '.';
	    name_in_entry (deh + 1, 1)[1] = '.';
	    dirty = 1;
	}
    }


    end = item + ih_item_len (ih);
    deh += ih_entry_count (ih);
    entries = (char *)deh;
    total_entry_len = ih_item_len (ih) - DEH_SIZE * ih_entry_count (ih);
    i = ih_entry_count (ih);

    bad_entries = 0;
    do {
	i --;
	deh --;
	masked_offset = deh_offset (deh);
	name_len = prob_name (fs, &entries, total_entry_len, masked_offset);
	if (!name_len) {
	    if (!bad_entries) {
		deh->deh_location = cpu_to_le16 (entries - item);
	    } else {
		deh->deh_location = cpu_to_le16 (deh_location (deh + 1) + 1);
	    }
	    bad_entries ++;

	    /*fsck_log ("verify_directory_item: entry %d: in string \'%s\' there is no substring matching hash %ld\n",
	      i, bad_name (entries, total_entry_len), masked_offset);*/
	    mark_de_bad (deh);
	    continue;
	}
	bad_entries = 0;
	/*fsck_log ("verify_directory_item: entry %d: found \"%s\" name matching hash %ld\n",
	  i, bad_name (entries, name_len), masked_offset);*/

	/* 'entries' points now to the name which match given offset - so, set deh_location */
	deh->deh_location = cpu_to_le16 (entries - item);
	deh->deh_state = 0;
	mark_de_visible (deh);

	entries += name_len;
	total_entry_len = end - entries;
	/* skip padding zeros */
	while (!*entries) {
	    entries ++;
	    total_entry_len --;
	}
	/* 'entries' points now at the place where next (previous) entry
           should start */
    } while ((char *)deh != item);


    /* fixme: this will not work if all entries are to be deleted */
    for (i = 0; i < ih_entry_count (ih); i ++, deh ++) {
	deh = (struct reiserfs_de_head *)B_I_PITEM (bh, ih) + i;
	if (de_bad (deh)) {
	    cut_entry (fs, bh, item_num, i);
	    i --;
	}
/*
	    fsck_log ("verify_directory_item: %d-th entry is to be deleted: "
		      "\"%s\" does not match to hash %lu\n", 
		      i, bad_name (name_in_entry (deh, i), name_length (ih, deh, i)),
		      deh_offset (deh));
*/
    }

    /* item looks ok */
    return 0;
}


/* do this on pass 0 with every leaf marked used */

/* FIXME: we can improve fixing of broken keys: we can ssfe direct items which
   go after stat data and have broken keys */
static void handle_leaf (reiserfs_filsys_t fs, struct buffer_head * bh)
{
    int i, j;
    struct item_head * ih;
    __u32 * ind_item;
    unsigned long unfm_ptr;
    int dirty = 0;

 start_again:

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih ++) {
	/* sometimes stat datas get k_objectid==0 or k_dir_id==0 */
	if (ih->ih_key.k_dir_id == 0 || ih->ih_key.k_objectid == 0) {
	    if (i == (B_NR_ITEMS (bh) - 1)) {
		if (i == 0) {
		    fsck_log ("pass0: item %k is alone in the block %lu\n",
			      &ih->ih_key, bh->b_blocknr);
		    return;
		}
		/* delete last item */
		delete_item (fs, bh, i - 1);
		dirty = 1;
		return;
	    }
	    /* there is next item: if it is not stat data - take its k_dir_id
               and k_objectid. if key order will be still wrong - the changed
               item will be deleted */
	    if (!is_stat_data_ih (ih + 1)) {
		ih->ih_key.k_dir_id = (ih + 1)->ih_key.k_dir_id;
		ih->ih_key.k_objectid = (ih + 1)->ih_key.k_objectid;
		set_offset (KEY_FORMAT_1, &ih->ih_key, 0);
		set_type (KEY_FORMAT_1, &ih->ih_key, TYPE_STAT_DATA);
		fsck_log ("pass0: block %lu: item %d had zero k_dir_id or k_objectid - fixed to %k\n",
			  bh->b_blocknr, i, &ih->ih_key);
		dirty = 1;
	    }
	}

	if (i &&  comp_keys (&(ih - 1)->ih_key, &ih->ih_key) != -1) {
	    /* previous item has key not smaller than the key of currect item */
	    if (is_stat_data_ih (ih - 1) && !is_stat_data_ih (ih)) {
		/* fix key of stat data such as if it was stat data of that item */
		fsck_log ("pass0: block %lu: %d-th item %k is out of order, made a stat data of %d-th (%k)\n",
			  bh->b_blocknr, i - 1, &(ih - 1)->ih_key, i, &ih->ih_key);
		(ih - 1)->ih_key.k_dir_id = ih->ih_key.k_dir_id;
		(ih - 1)->ih_key.k_objectid = ih->ih_key.k_objectid;
		set_offset (KEY_FORMAT_1, &(ih - 1)->ih_key, 0);
		set_type (KEY_FORMAT_1, &(ih - 1)->ih_key, TYPE_STAT_DATA);
		dirty = 1;
	    } else {
		/* ok, we have to delete one of these two - decide which one */
		int retval;

		/* something will be deleted */
		dirty = 1;
		retval = upper_correct (bh, ih - 1, i - 1);
		switch (retval) {
		case 0:
		    /* delete upper item */
		    fsck_log ("pass0: block %lu: %d-th (upper) item (%k) is out of order, deleted\n",
			      bh->b_blocknr, i - 1, &(ih - 1)->ih_key);
		    delete_item (fs, bh, i - 1);
		    goto start_again;

		case 1:
		    /* delete lower item */
		    fsck_log ("pass0: block %lu: %d-th (lower) item (%k) is out of order, deleted\n",
			      bh->b_blocknr, i, &ih->ih_key);
		    delete_item (fs, bh, i);
		    goto start_again;

		default:
		    /* upper item was the first item of a node */
		}

		retval = lower_correct (bh, ih, i);
		switch (retval) {
		case 0:
		    /* delete lower item */
		    fsck_log ("pass0: block %lu: %d-th (lower) item (%k) is out of order, deleted\n",
			      bh->b_blocknr, i, &ih->ih_key);
		    delete_item (fs, bh, i);
		    goto start_again;

		case 1:
		    /* delete upper item */
		    fsck_log ("pass0: block %lu: %d-th (upper) item (%k) is out of order, deleted\n",
			      bh->b_blocknr, i - 1, &(ih - 1)->ih_key);
		    delete_item (fs, bh, i - 1);
		    goto start_again;

		default:
		    /* there wer only two items in a node, so we could not
                       decide what to delete, go and ask user */
		}
		fsck_log ("pass0: which of these items looks better (other will be deleted)?\n"
			  "%h\n%h\n", ih - 1, ih);
		if (fsck_user_confirmed (fs, "1 or 2?", "1\n", 1))
		    delete_item (fs, bh, i - 1);
		else
		    delete_item (fs, bh, i);
		goto start_again;
	    }
	}

	if (is_direntry_ih (ih)) {
	    verify_directory_item (fs, bh, i);
	    continue;
	}


	/* FIXME: should we do something with other items? */
	if (!is_indirect_ih (ih))
	    continue;
	
	ind_item = (__u32 *)B_I_PITEM (bh, ih);
	for (j = 0; j < I_UNFM_NUM (ih); j ++) {
	    unfm_ptr = le32_to_cpu (ind_item [j]);
	    if (!unfm_ptr)
		continue;
	    
	    if (fsck_mode (fs) == FSCK_ZERO_FILES) {
		ind_item [j] = 0;
		reiserfs_bitmap_clear_bit (fsck_new_bitmap(fs), unfm_ptr);
		tmp_zeroed ++;
		dirty = 1;
		continue;
	    }

	    if (not_data_block (fs, unfm_ptr) || /* journal area or bitmap or super block */
		unfm_ptr >= SB_BLOCK_COUNT (fs)) {/* garbage in pointer */

		stat_wrong_pointer_found (fs);
		fsck_log ("pass0: %d-th pointer (%lu) in item %k (leaf block %lu) is wrong\n",
			  j, unfm_ptr, &ih->ih_key, bh->b_blocknr);
		continue;
	    }
	    if (!was_block_used (unfm_ptr)) {
	      /* this will get to a pool of allocable blocks */
	      ind_item [j] = 0;
	      dirty = 1;
	      stat_wrong_pointer_found (fs);
	      continue;
	    }

	    /* mark block in bitmaps of unformatted nodes */
	    register_unfm (unfm_ptr);
	}
    }

    if (B_NR_ITEMS (bh) < 1)
	die ("pass0: block %lu got all items deleted\n",
	     bh->b_blocknr);

    pass0_mark_leaf (bh->b_blocknr);

    if (dirty) {
	stat_pass0_leaf_corected (fs);
	mark_buffer_dirty (bh);
    }
}







/*############### end of code for debugging ###########################################*/



/* this tests  */
static void do_test (reiserfs_filsys_t fs, unsigned long block)
{
    struct buffer_head * bh;

    fprintf (stderr, "=========== BLock %lu ============\n", block);
    bh = bread (fs->s_dev, block, fs->s_blocksize);
    if (!bh)
	die ("do_test: bread failed");
/*    print_block (bh, 3, -1, -1);*/

    /* function to test */handle_leaf (fs, bh);

    fprintf (stderr, "\n=====================================================\n");
    print_block (fs, bh, 3, -1, -1);
    brelse (bh);
}


/* FIXME: need to open reiserfs filesystem first */
int main (int argc, char * argv[])
{
    char * file_name;
    reiserfs_filsys_t fs;
    int error;

    print_banner ("debugreiserfs");
 
    file_name = parse_options (argc, argv);

    fs = reiserfs_open (file_name, O_RDONLY, &error, 0);
    if (!fs) {
	fprintf (stderr, "\n\ndumpreiserfs: can not open reiserfs on \"%s\": %s\n\n",
		 file_name, error ? strerror (error) : "there is no one");
	return 0;
    }

    switch (mode) {
    case DO_PACK:
	do_pack (fs);
	break;

    case DO_CORRUPT:
	reiserfs_reopen (fs, O_RDWR);
	do_corrupt_one_block (fs, opt_block_to_corrupt);
	break;

    case DO_DUMP:
	do_dump_tree (fs);
	break;

    case DO_SCAN:
    case DO_SCAN_FOR_NAME:
	do_scan (fs);
	break;

    case DO_RECOVER:
	do_recover (fs);
	break;

    case DO_TEST:
    {
	int i;
#if 0
	int arr[] = {53033, 179201, 844702,
		     844913,
		     877768,
		     879067,
		     907631,
		     925323,
		     2241275,
		     2241343,
		     2241397,
		     2241511,
		     2241553,
		     2241635,
		     2241644,
		     2241654,
		     2241711,
		     2241721,
		     2241727,
		     2241740,
		     2241762,
		     2241766,
		     2241770,
		     2241812,
		     2241820,
		     2241827,
		     2241831,
		     2241878,
		     2241886,
		     2241971
	};
#endif
	/* blocks containing broken directory items on Joop's filesystem */
	int arr[] = {/*53033,*/ 838396/*, 1597036*//*, 1919589, 2715962*/};
	
	
	for (i = 0; i < sizeof (arr) / sizeof (arr[0]); i ++)
	    do_test (fs, arr[i]);
	break;
    }
    }

    reiserfs_close (fs);
    return 0;
}
